
########################################################################
# generic pass-through intializer generator.
# supplies "init" subroutine to merge metadata via EVERY::LAST call
# in constructor.
########################################################################

########################################################################
# housekeeping
########################################################################

package NEXT::init;

our $VERSION = '0.99';

use strict;
use warnings;

use Symbol;

# debug output

use Data::Dumper;

########################################################################
# package variables
########################################################################

########################################################################
# subroutines
########################################################################

########################################################################
# utility subs to handle data on @_.
# re-used in import and init.
#
# lexicals avoids unintended inheritence.

########################################################################
# package variables
########################################################################

my @switchez = qw( type );

our $verbose = 0;

our $type = '';

########################################################################
# utility subs to handle data on @_.
# re-used in import and init.
#
# lexicals avoids accident inheritence.

my $switches =
sub 
{
	# reset the standard switches.

	for( @switchez )
	{
		my $a = *{qualify_to_ref $_, __PACKAGE__}{SCALAR};

		$$a = 0;
	}

	while( @_ && $_[0] =~ /^:(\S+)$/ )
	{
		my ( $name, $value ) = split /=/, $1;

		$value ||= 1;

		my $a =  *{qualify_to_ref $name, __PACKAGE__}{SCALAR};

		$$a = uc $value;

		shift;
	}

	()
};

# first time around it takes a bit more work to 
# determine the proper type for an object. after
# the metadata and init sub's have been installed
# this doesn't need to be done again.

my $objtype =
sub
{
	# if the arguments are a referent then figure out what
	# they are and use that as a default for the type if
	# one was not passed in via ":type=blah".
	#
	# caller gets back true for array (hash is the default).
	#
	# eval's handle blessed referents.

	$type ||= 
	do
	{
		if( my $a = ref $_[0] )
		{ 
			if( $a =~ /^(HASH|ARRAY)$/ )
			{
				$1
			}
			elsif( eval{ %{$_[0]} } )
			{
				'HASH'
			}
			elsif( eval{ @{$_[0]} } )
			{
				'ARRAY'
			}
			else
			{
				die "Unusable referent:\n" . Dumper $_[0];
			}
		}
		else
		{
			'HASH'
		}
	};

	# caller can check $type for the exact type, main 
	# distinction at this point is whether to use the
	# default of a hash.

	$type eq 'HASH'
};

########################################################################
# catch: you can't check for ref $x eq 'HASH' since the reference
# may be blessed already...
#
# fix: eval the expansion in various formats and then give up.
#
# note: this has to be called AFTER switches, above, so that
# the stack has real data on it...
#
# all this does is expand a referent into a flat list.

my $objvaluz =
sub
{
	my @a = ();

	if( defined eval { @a = @{ $_[0] } } )
	{
		@a
	}
	elsif( defined eval { @a = %{ $_[0] } } )
	{
		@a
	}
	else
	{
		@a = @_
	};
};

###########################################################
# passed class meta-data as a list or referent. lists 
# become hashes; others are left as-is (only arrays and
# hashes are handled gracefully at this point).
#
# the metadata is installed into the package as %meta or 
# @meta, along with a closure as &init to merge the 
# metadata with an exsiting object's contents.
#
# note: mixing metdata types (e.g., arrays with hashes) is
# currently not supported.
#
# this also updates the callers @ISA to include init for
# if the caller cannot 'construct'.
#
##			push @{ $_[0] }, @$ref;
#
##			exists $_[0]->{$_} or $_[0]->{$_} = $ref->{$_}
##				for keys %$ref;

sub import
{
	local $\ = "\n";
	local $, = ", ";

	# discard the current package name.

	shift;

	my $caller = caller;

	print STDERR "\nProcessing: $caller\n" . Dumper \@_
		if $verbose;

	&$switches;

	my $usehash  = &$objtype;

	my @objvaluz = &$objvaluz;

	# i.e., merge the current metadata into the object
	# via hash-slice. this works with EVERY::LAST to 
	# allow more-derived classes to overwrite data from
	# less-derived ones.

	my ( $ref, $sub ) = 
	do
	{
		my( $a, $b ) = (0, 0);

		if( $usehash )
		{
			$a = { @objvaluz };
			$b = sub { @{$_[0]}{ keys %$a } = values %$a }
		}
		else
		{
			$a = \@objvaluz;
			$b = sub { unshift @{ $_[0] }, @$a };
		}

		( $a, $b )
	};

	# now install the subroutine in the caller's 
	# namespace and ensure that it can find a 
	# constructor (effectively converts the use
	# to a use base with arguments).
	#
	# installing %meta allows the caller to mangle
	# the values after initialzing them.

	my $ISA  = *{qualify_to_ref 'ISA',  $caller}{ARRAY};
	my $init =   qualify_to_ref 'init', $caller;
	my $meta =   qualify_to_ref 'meta', $caller;

	unshift @$ISA, __PACKAGE__;

	*$meta = $ref;

	*$meta = sub { $ref };

	*$init = $sub;

	$DB::single = 1;

	# someplace to hang a breakpoint

	1
}

########################################################################
# this may be called as a class method or via a
# factory object.
#
# if this is passed a referent the data is used as-is,
# otherwise @_ is copied into an anon hash. mis-matched
# metadata in the init calls is fatal.
#
# after constructing an object from the data 
# initialize it within all of the classes 
# its class inherits from via NEXT.
#
# hash objects can request before-and-after dumper output
# by setting $obj->{verbose} true; arrays are silent.
#
# hashes can pass in a true value for $obj->{verbose}
# for before-and-after dumper output of the data;
# arrays are slient.
#
# catch: the unshift in array handling leaves the
# arguments at the END, probably not where anyone
# wants them... fix is to start with an empty 
# object, populate it, then add the arguments in 
# at the end. alternative is to use use push on 
# arrays and assign-if-not-exsists on the hashes
# instead of a hash-slice assignment...

sub construct
{
	use NEXT;

	my $item = shift;

	my $class = ref $item || $item;

	# the caller may choose to override the object
	# type -- for example to re-use classes with 
	# mismatched data types. in any case, strip
	# the switches and extract the remaining values
	# as add'l values for the new object.

	&$switches;

	my @objvaluz = &$objvaluz;

	# if there is no data in the class then use the 
	# meta sub to get a base class type.

	$type ||= ref $class->meta
		or die "Unable to determine data type for: $class";

	my $obj = $type eq 'HASH'? {} : [];

	bless $obj, $class;

	$obj->EVERY::LAST::init;

	if( $type eq 'HASH' )
	{
		my %argz = @objvaluz;

		@{$obj}{ keys %argz } = values %argz;
	}
	else
	{
		unshift @$obj, @objvaluz;
	}

	$obj
}

# keep require happy 

1

__END__

=head1 NAME

Class::init

Initialize objects using metadata, exports 'init' sub to handle
inherited data, 'construct' method to do the deed.

=head1 SYNOPSIS

=head2 hash-based metadata

Each level in the init does an array-slice assignment, 
overwriting values from less-derived classes as necessary.

	#!/blah/perl

	use Class::init;

	# use object metadata as is or add data to it,
	# which may overwrite the class data.

	my $obj1 = Foo->construct();

	my $obj2 = Foo->construct( additional => metadata );

	my $obj3 = Bar->construct;

	my $obj4 = Foo->construct( yetmore => meta );

	package Foo;

	use Class::init
	(
		this => 'that',
		that => 'another',
		...
	);

	pakcage Bar;
	use base qw( Foo );

	# result:
	# ( this => 'whatever', foo => 'bar', 'that => another' )

	use Class::init
	{
		# this overwrites 'Foo' base class' value.

		this => 'whatever',

		foo  => 'Bar',
	};


=head2 array-based metadata

This unshifts the data from each derived class onto the 
metadata. Result is a queue (via pop) or stack (via shift).
Metadata must be supplied as an array referent.

	#!/blah/perl

	use Class::init;

	# use object metadata as is or add data to it,
	# which may overwrite the class data.

	my $handlerz = Bar->construct;

	$_->( $data ) for @$handlerz;

	my $more_handlerz =
		Bar->construct( :array sub { print 'Foo Wins!' } );

	package Foo;

	use Class::init
	[
		sub { print "Foo-1" },
		sub { print "Foo-2" },
	];

	pakcage Bar;
	use base qw( Foo );

	use Class::init
	[
		sub { print "Bar-1" },
		sub { print "Bar-2" },
	];

Using the ":array" metadata the following three calls 
are equivalent:

	my $a = MyArray->construct( ':array',   qw( e f g h ) );
	my $b = MyArray->construct( ':array', [ qw( e f g h ) ] );
	my $c = MyArray->construct( ':array', { qw( e f g h ) } );

All will give back an array referent as the object, with
qw ( e f g h ) as the first members followed by whatever
MyArrary defines/inherits.  Without the leading ":array"
argument the first and last examples ($a and $c) would
have returned a hash.

=head1 DESCRIPTION

This is a generic initializer class for objects that need
to inherit hash or array data. When use-ed the import module
installs an "init" method which merges the current data into
the object calling the method, an array/hash named 'meta' (in
case the class needs to modify its data after initialization),
and unshifts NEXT::init onto @ISA (i.e., an implied use base).

The class supplies a "construct" method which creates a new
object then calls "$obj->EVERY::LAST::init", which daisy-chains 
the object through its installed init handlers.

For effeciency the init handers are closures which handle
either hash or array data propery (i.e., no repeated if-
logic, changing the base data type of the object after
construction will likely break the associated init handlers).

The object's type (hash or array) is determined either by
the type of referent passed into init or an initial ":array"
argument; lacking a referent and the ":array" argument the
type defaults to a hash.

From test.pl:

	my $hash0 = foo->construct( %baseline_hash );
	my $hash1 = bar->construct( %baseline_hash, dbi_connect => \@db1 );
	my $hash2 = foo_plus->construct( %baseline_hash, dbi_connect => \@db2 );
	my $hash3 = foo_plus_bar->construct( \%baseline_hash );


	my $arry4 = MyArray->construct( [] );
	my $arry5 = MyArray->construct( [ qw( e f g h ) ] );
	my $arry6 = MyArray->construct( ':array', [ qw( e f g h ) ] );
	my $arry6 = MyArray->construct( ':array', { qw( e f g h ) } );
	
Note that $arry6 uses a hash referent but requests an array,
at which point the hash is listified into an array which is
then unshifted. This can be marginally useful for reverse-
inheritence via:

	my $obj = Foo->construct( ':array' );

	%array = %$obj;

which will give priority in values to the least-derived
class (which has later data on the array).

Hash data is inherited via EACH::init:

	exists $obj->{$_} or $obj->{$_} = $meta->{$_}
		for keys %$meta;

Array data via:

	push @$obj, @$meta;

This was done because 

One example use is an inherited handler stack:

	my $handler1 = Baseline->construct( [ sub { top handler } ] );

	my $handler2 = MoreHandleras->construct( [] );

	package MoreHandlers;

	use base( Baseline );

	use NEXT::init
	[
		sub { ...  More 1 ... }
		sub { ...  More 2 ... },
	];

	package Baseline;

	use NEXT::init
	[
		sub { ... Baseline 1 ... },
		sub { ... Baseline 2 ... },
	];

Leaves $handler1 with:

	[
		sub { top handler },
		sub { ... Baseline 1 ... },
		sub { ... Baseline 2 ... },
	];

and $handler2 with:

	use NEXT::init
	[
		sub { ...  More 1 ... }
		sub { ...  More 2 ... },
		sub { ... Baseline 1 ... },
		sub { ... Baseline 2 ... },
	];


Another is a query engine that inherits query data and
takes its dbi connect as an argument at construction time:

	my $lk = Lookup->construct( dbi_args => [ ... ] );

	pacakge Lookup;

	use base qw( BaselineQueries );

	use NEXT::init
	(
		query1 => q{ ... },
		query2 => q{ ... },
	);

Will create an object $lk with the contents of BaselineQueries 
overridden by Lookup, with added arguments "dbi_args".

One use of this is constructing more complicated data
structures where merging the values piecemeal can be a
pain:

	my $messy_struct =
	{
		querys => Lookup->construct( ... ),

		handlers => MoreHandlers->construct( ... ),
	};

=head2 %meta and @meta

The values passed into NEXT::init via use are installed
as *$class::meta = ref. This allows the class to inspect
its values or modify them over time. One use would be to
fold command line arguments into boilerplate values before
values are created from it (e.g., for an object that is
used with DBI). 

=head1 AUTHOR

Steven Lembark (lembark@wrkhors.com)

=head2 To Do

- Find a graceful way of merging/skipping mis-matched
  base types in the init chain (e.g., trying to merge
  hash and array data). 

- Forcing the data type to match the type of inherited
  data automatically (current pratice is an execption
  when the mis-matched access is encountered).

- Deal with non-array/-hash referent types (suggestions
  welcome). CODE ref's might be executed in the proper
  order but that doesn't seem much like "inheritence".

=head1 COPYRIGHT

This code is released as-is under the same terms as Perl-5.8 (or
any later version of perl the user preferrs).

=head1 SEE ALSO

NEXT perlreftut perlootut
